@@ -28,7 +28,7 @@ limitations under the License.
 
 __author__ = 'Marcin Ulikowski'
 __email__ = 'marcin@ulikowski.pl'
-__version__ = '2022.11.0'
+__version__ = '2022.12.0'
 
 
 import sys
@@ -37,7 +37,6 @@ import re
 import signal
 import time
 import socket
-import random
 import itertools
 import getopt
 import concurrent.futures
@@ -63,21 +62,6 @@ else:
 	cYEL = cCYA = cBLU = cGRN = cRST = sBRI = sRST = sCLR = ''
 
 
-class DNSSet(set):
-	def __init__(self):
-		super(set, self).__init__()
-
-	def errors(self):
-		return {item for item in self if item._error}
-
-	def answered(self):
-		return {item for item in self if not item._error}
-
-	def empty(self):
-		return {item for item in self
-			if not bool(item.a or item.cname or item.ns or item.servfail or item.refused)}
-
-
 class DNS(dict):
 	'''Simplified and flattened DNS response object'''
 
@@ -91,7 +75,6 @@ class DNS(dict):
 	def __init__(self, domain=''):
 		super(dict, self).__init__()
 		self.domain = domain
-		self._error = False
 		self.servfail = False
 		self.refused = False
 		self.a = []
@@ -152,10 +135,7 @@ class QResolver():
 
 	@staticmethod
 	def _build_query(fqdn, rdtype=RDTYPE_A):
-		j = lambda x: chr(len(x)) + x
-		l = fqdn.split('.')
-		qname = ''.join(map(j, l)).encode()
-
+		qname = b''.join([len(x).to_bytes(1, 'big') + x for x in fqdn.encode('idna').split(b'.')])
 		return b''.join([b'\x19\x86\x01\x00\x00\x01\x00\x00\x00\x00\x00\x01',
 			qname,
 			b'\x00',
@@ -183,7 +163,7 @@ class QResolver():
 				break
 			else:
 				label = data[i+1: i+1+b]
-				labels.append(label.decode())
+				labels.append(label.decode('idna'))
 				i += b
 
 			i += 1
@@ -202,20 +182,24 @@ class QResolver():
 		while True:
 			rl = reader.read(1)[0]
 			if rl:
-				rn.append(reader.read(rl).decode())
+				rn.append(reader.read(rl))
 			else:
 				reader.skip(4)
 				break
 
-		rname = '.'.join(rn)
+		rname = b'.'.join(rn).decode('idna')
 		if qname != rname:
 			raise DNSException('Inconsistent DNS query and response: {} <> {}'.format(qname, rname))
 
 		flags = to_int(header[2:4])
 		rcode = flags & 0x0003
 
+		answer_rr = to_int(header[6:8])
+
 		if rcode == cls.RCODE_NXDOMAIN:
-			raise NXDOMAIN('Domain name does not exist: {}'.format(qname))
+			# ignore NXDOMAIN if answer is present
+			if answer_rr == 0:
+				raise NXDOMAIN('Domain name does not exist: {}'.format(qname))
 
 		elif rcode == cls.RCODE_SERVFAIL:
 			raise SERVFAIL('Server failed to complete request: {}'.format(qname))
@@ -235,8 +219,6 @@ class QResolver():
 			cls.RDTYPE_PTR: 'ptr',
 			}
 
-		answer_rr = to_int(header[6:8])
-
 		for _ in range(answer_rr):
 			reader.skip(2)
 			type_id = to_int(reader.read(2))
@@ -283,8 +265,7 @@ def nslookup(resolver, fqdn):
 	try:
 		res = QResolver.resolve(resolver, fqdn, rdtype=QResolver.RDTYPE_A)
 	except NXDOMAIN:
-		# discard non-existing domain
-		return {}
+		raise
 	except SERVFAIL:
 		dns.servfail = True
 		return dns
@@ -292,8 +273,7 @@ def nslookup(resolver, fqdn):
 		dns.refused = True
 		return dns
 	except Exception:
-		dns._error = True
-		return dns
+		raise
 	else:
 		dns.a = sorted(res.get('a', []))
 		dns.cname = res.get('cname', [])
@@ -347,6 +327,8 @@ class Fuzzer():
 	def _fuzz(self, word):
 		swaps = self.swaps
 		words = list(filter(None, word.split('-')))
+		for i in range(1, len(words)):
+			yield '-'.join(words[:i])
 		for i, w in enumerate(words):
 			pre = words[:i]
 			suf = words[i+1:]
@@ -382,14 +364,13 @@ class Fuzzer():
 			else:
 				self._prep(item)
 
-
 	def mutations(self):
 		for sub in self.subdomains:
 			yield '.'.join([sub, self.domain])
 
 	def wildations(self):
-		for wild in {'.'.join(['*'] + sub.split('.')[1:]) for sub in self.subdomains}:
-			yield '.'.join([wild, self.domain])
+		for wild in {'.'.join(['*'] + sub.split('.', 1)[1:]) for sub in self.subdomains}:
+			yield wild + '.' + self.domain
 
 	def count(self):
 		return len(self.subdomains)
@@ -405,11 +386,8 @@ def chunkify(iterable, chunk=1000):
             return
 
 
-def wild_cmp(a, b):
-	for x, y in itertools.zip_longest(a.split('.')[::-1], b.split('.')[::-1]):
-		if x != y and x != '*' and y != '*':
-			return False
-	return True
+def to_wild(domain):
+	return '.'.join(['*', domain.partition('.')[2]])
 
 
 def status(message):
@@ -469,7 +447,7 @@ def run():
 
 	if not domain:
 		fatal('domain name is required')
-	args.domain = domain[0].encode('idna').decode()
+	args.domain = domain[0]
 
 	for opt, val in opts:
 		if opt == '--input':
@@ -480,6 +458,8 @@ def run():
 					fatal('unable to open {} ({})'.format(val, err.strerror.lower()))
 
 		elif opt == '--output':
+			if os.path.exists(val):
+				fatal('{} already exists'.format(val))
 			try:
 				args.output = open(val, 'w', encoding='utf-8')
 			except OSError as err:
@@ -544,21 +524,27 @@ def run():
 			except UnicodeDecodeError:
 				fatal('unable to parse {}'.format(f.name))
 
-		randnx = 'subzuf-{:x}.com'.format(random.getrandbits(64))
+		randnx = 'subzuf-{:x}.com'.format(int.from_bytes(os.urandom(8), 'little'))
 
-		futures = {executor.submit(nslookup, resolver=resolv, fqdn=randnx): resolv
+		futures = {executor.submit(QResolver.resolve, resolver=resolv, fqdn=randnx, timeout=1): resolv
 			for resolv in resolvers}
 
 		for future in concurrent.futures.as_completed(futures):
-			res = future.result()
-			resolv = futures.pop(future)
+			try:
+				res = future.result()
+			except NXDOMAIN:
+				futures.pop(future)
+				continue
+			except Exception:
+				pass
 
-			if res:
-				resolvers.remove(resolv)
-				status('rejected resolver: {}\n'.format(resolv))
+			resolv = futures.pop(future)
+			resolvers.remove(resolv)
+			status('rejected resolver: {}\n'.format(resolv))
 
 	if not resolvers:
 		fatal('all DNS resolvers have been rejected')
+	iresolver = itertools.cycle(resolvers)
 
 	# read input data
 
@@ -573,9 +559,9 @@ def run():
 
 	# all checks complete, input cases ready, start main loop
 
-	subdomains = DNSSet()
-
+	subdomains = set()
 	wildcards = dict()
+	errors = 0
 
 	s_perf = time.perf_counter()
 
@@ -591,17 +577,22 @@ def run():
 		if args.wildcard in ('filter', 'reject'):
 
 			for wbucket in chunkify(fuzz.wildations(), args.threads*100):
-				futures = {executor.submit(nslookup, resolver=random.choice(resolvers), fqdn=wtest): wtest
+				futures = {executor.submit(nslookup, resolver=next(iresolver), fqdn=wtest): wtest
 					for wtest in wbucket if wtest not in wildcards}
 	
 				for future in concurrent.futures.as_completed(futures):
-					wtest = future.result()
-					domain = futures.pop(future)
+					try:
+						wtest = future.result()
+					except Exception:
+						pass
+					else:
+						if not (wtest.servfail or wtest.refused):
+							status('wildcard found: {}\n'.format(wtest.domain))
+							wildcards[wtest.domain] = wtest.a
+					finally:
+						domain = futures.pop(future)
+
 					status('wildcards: {} | testing: {}'.format(len(wildcards), domain))
-	
-					if wtest:
-						status('wildcard found: {}\n'.format(domain))
-						wildcards[domain] = wtest.a
 
 		status('scanning...')
 
@@ -610,7 +601,7 @@ def run():
 		f_cnt = fuzz.count()
 		p_perf = time.perf_counter()
 
-		exist = DNSSet()
+		exist = set()
 
 		for bucket in chunkify(fuzz.mutations(), args.threads*100):
 
@@ -622,34 +613,37 @@ def run():
 						done += 1
 						continue
 
-				future = executor.submit(nslookup, resolver=random.choice(resolvers), fqdn=elem)
+				future = executor.submit(nslookup, resolver=next(iresolver), fqdn=elem)
 				futures[future] = elem
 
 			for future in concurrent.futures.as_completed(futures):
-				res = future.result()
-				domain = futures.pop(future)
-
-				done += 1
-
-				if res:
+				try:
+					res = future.result()
+				except NXDOMAIN:
+					pass
+				except Exception:
+					errors += 1
+				else:
 					if args.wildcard == 'filter':
-						if not (res.a in wildcards.values() and any([wild_cmp(res.domain, wild) for wild in wildcards])):
+						if wildcards.get(to_wild(res.domain)) != res.a:
 							exist.add(res)
 					else:
 						exist.add(res)
+				finally:
+					futures.pop(future)
+					done += 1
 
 				if done > args.threads and done % args.threads:
 					continue
 
 				rate = int(done / (time.perf_counter() - p_perf)) + 1
 				eta = int((f_cnt - done) / rate)
-				errperc = len(exist.errors()) / done
 
 				status(' | '.join([
 					'loop: #{}'.format(loopn),
 					'progress: {}{:.2%}{}'.format(sBRI, done / f_cnt, sRST),
-					'found: {}{}{}/{}'.format(sBRI, len(exist.answered()), sRST, f_cnt),
-					'errors: {:.2%}'.format(errperc),
+					'found: {}{}{}/{}'.format(sBRI, len(exist), sRST, f_cnt),
+					'errors: {} ({:.2%})'.format(errors, errors / done),
 					'speed: {}/s'.format(rate),
 					'eta: {:d}h {:02d}m {:02d}s'.format(eta // 3600, eta % 3600 // 60, eta % 3600 % 60),])
 					)
@@ -662,7 +656,7 @@ def run():
 
 		extras = set()
 
-		for item in exist.answered():
+		for item in exist:
 			for i in item.cname + item.ns + item._ptr:
 				if i.endswith('.' + args.domain):
 					extras.add(i)
@@ -680,14 +674,14 @@ def run():
 
 	# display summary and results
 
-	results = sorted(subdomains.answered())
+	results = sorted(subdomains)
 
 	ttime = int(time.perf_counter() - s_perf)
 
 	status(' | '.join([
 		'complete',
 		'found: {}{}{}'.format(sBRI, len(results), sRST),
-		'errors: {}{}{}'.format(sBRI, len(subdomains.errors()), sRST),
+		'errors: {}{}{}'.format(sBRI, errors, sRST),
 		'time: {:d}h {:02d}m {:02d}s'.format(ttime // 3600, ttime % 3600 // 60, ttime % 3600 % 60),])
 		+ '\n\n'
 		)
